@@ -25,7 +25,9 @@ const Convenience = Me.imports.convenience;
 const appSys = Shell.AppSystem.get_default();
 
 const APPLICATION_ICON_SIZE = 32;
+const HORIZ_FACTOR = 5;
 const MENU_HEIGHT_OFFSET = 132;
+const NAVIGATION_REGION_OVERSHOOT = 50;
 
 const ActivitiesMenuItem = new Lang.Class({
     Name: 'ActivitiesMenuItem',
@@ -53,12 +55,21 @@ const ApplicationMenuItem = new Lang.Class({
 	this._app = app;
         this._button = button;
 
-        let icon = this._app.create_icon_texture(APPLICATION_ICON_SIZE);
-	this.addActor(icon);
+        this._iconBin = new St.Bin();
+        this.addActor(this._iconBin);
 
         let appLabel = new St.Label({ text: app.get_name() });
         this.addActor(appLabel, { span: -1, expand: true });
         this.actor.label_actor = appLabel;
+
+        let textureCache = St.TextureCache.get_default();
+        let iconThemeChangedId = textureCache.connect('icon-theme-changed',
+                                                      Lang.bind(this, this._updateIcon));
+        this.actor.connect('destroy', Lang.bind(this,
+            function() {
+                textureCache.disconnect(iconThemeChangedId);
+            }));
+        this._updateIcon();
     },
 
     activate: function(event) {
@@ -76,6 +87,10 @@ const ApplicationMenuItem = new Lang.Class({
 
     _getPreferredWidth: function(actor, forHeight, alloc) {
         alloc.min_size = alloc.natural_size = -1;
+    },
+
+    _updateIcon: function() {
+        this._iconBin.set_child(this._app.create_icon_texture(APPLICATION_ICON_SIZE));
     }
 });
 
@@ -88,6 +103,9 @@ const CategoryMenuItem = new Lang.Class({
 	this._category = category;
         this._button = button;
 
+        this._oldX = -1;
+        this._oldY = -1;
+
         let name;
         if (this._category)
             name = this._category.get_name();
@@ -95,6 +113,7 @@ const CategoryMenuItem = new Lang.Class({
             name = _("Favorites");
 
         this.addActor(new St.Label({ text: name }));
+        this.actor.connect('motion-event', Lang.bind(this, this._onMotionEvent));
     },
 
     activate: function(event) {
@@ -103,6 +122,81 @@ const CategoryMenuItem = new Lang.Class({
 	this.parent(event);
     },
 
+    _isNavigatingSubmenu: function([x, y]) {
+        let [posX, posY] = this.actor.get_transformed_position();
+
+        if (this._oldX == -1) {
+            this._oldX = x;
+            this._oldY = y;
+            return true;
+        }
+
+        let deltaX = Math.abs(x - this._oldX);
+        let deltaY = Math.abs(y - this._oldY);
+
+        this._oldX = x;
+        this._oldY = y;
+
+        // If it lies outside the x-coordinates then it is definitely outside.
+        if (posX > x || posX + this.actor.width < x)
+            return false;
+
+        // If it lies inside the menu item then it is definitely inside.
+        if (posY <= y && posY + this.actor.height >= y)
+            return true;
+
+        // We want the keep-up triangle only if the movement is more
+        // horizontal than vertical.
+        if (deltaX * HORIZ_FACTOR < deltaY)
+            return false;
+
+        // Check whether the point lies inside triangle ABC, and a similar
+        // triangle on the other side of the menu item.
+        //
+        //   +---------------------+
+        //   | menu item           |
+        // A +---------------------+ C
+        //              P          |
+        //                         B
+
+        // Ensure that the point P always lies below line AC so that we can
+        // only check for triangle ABC.
+        if (posY > y) {
+            let offset = posY - y;
+            y = posY + this.actor.height + offset;
+        }
+
+        // Ensure that A is (0, 0).
+        x -= posX;
+        y -= posY + this.actor.height;
+
+        // Check which side of line AB the point P lies on by taking the
+        // cross-product of AB and AP. See:
+        // http://stackoverflow.com/questions/3461453/determine-which-side-of-a-line-a-point-lies
+        if (((this.actor.width * y) - (NAVIGATION_REGION_OVERSHOOT * x)) <= 0)
+             return true;
+
+        return false;
+    },
+
+    _onMotionEvent: function(actor, event) {
+        if (!Clutter.get_pointer_grab()) {
+            this._oldX = -1;
+            this._oldY = -1;
+            Clutter.grab_pointer(this.actor);
+        }
+        this.actor.hover = true;
+
+        if (this._isNavigatingSubmenu(event.get_coords()))
+            return true;
+
+        this._oldX = -1;
+        this._oldY = -1;
+        this.actor.hover = false;
+        Clutter.ungrab_pointer();
+        return false;
+    },
+
     setActive: function(active, params) {
         if (active) {
             this._button.selectCategory(this._category, this);
@@ -231,6 +325,17 @@ const ApplicationsButton = new Lang.Class({
         return false;
     },
 
+    _onMenuKeyPress: function(actor, event) {
+        let symbol = event.get_key_symbol();
+        if (symbol == Clutter.KEY_Left || symbol == Clutter.KEY_Right) {
+            let direction = symbol == Clutter.KEY_Left ? Gtk.DirectionType.LEFT
+                                                       : Gtk.DirectionType.RIGHT;
+            if (this.menu.actor.navigate_focus(global.stage.key_focus, direction, false))
+                return true;
+        }
+        return this.parent(actor, event);
+    },
+
     _onVertSepRepaint: function(area) {
         let cr = area.get_context();
         let themeNode = area.get_theme_node();
@@ -319,7 +424,7 @@ const ApplicationsButton = new Lang.Class({
         this.leftBox = new St.BoxLayout({ vertical: true });
         this.applicationsScrollBox = new St.ScrollView({ x_fill: true, y_fill: false,
                                                          y_align: St.Align.START,
-                                                         style_class: 'vfade' });
+                                                         style_class: 'apps-menu vfade' });
         this.applicationsScrollBox.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
         let vscroll = this.applicationsScrollBox.get_vscroll_bar();
         vscroll.connect('scroll-start', Lang.bind(this, function() {
